"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.ProcessDispatcher = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }(); /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * @file {@link ProcessDispatcher} is simply a container of pre-forked {@link ModuleProcess} instances.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * It also a set of useful tools.
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * @author Denis Ilguzin
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      */

var _async = require("async");

var async = _interopRequireWildcard(_async);

var _utils = require("./utils");

var _moduleProcess = require("./moduleProcess");

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * <p><i><strong>Note: before running dispatching workflow do not forget to init listener on the module.</strong>
 * For that purpose, please invoke {@link ModuleProcess.listenIPCMessages} in the module root. It will engage module
 * to listen to IPC messages and expose module API available via sub-processing.</i></p>
 *
 * @param moduleName {string}   Module name (script file name). E.g. __filename
 * @param moduleOpts {Object}   Options object to pass to the module
 * @param logger {Object}
 * @class
 */
var ProcessDispatcher = exports.ProcessDispatcher = function () {

  /**
   * <p>Callback to run on function finished. <i>Note: It might contain arbitrary number of arguments (required at
   * least one - error)</i></p>
   *
   * @callback ProcessDispatcher~onReadyCallback
   * @param error {Error}   An Error object if the callback has raised an error.
   */

  /**
   *
   * @param moduleName
   * @param moduleOpts
   * @param logger
   */
  function ProcessDispatcher(moduleName, moduleOpts) {
    var logger = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : _utils.logger;

    _classCallCheck(this, ProcessDispatcher);

    this.moduleName = moduleName;
    this.moduleOpts = moduleOpts;
    this.availableProcesses = [];
    this.logger = logger;
  }

  /**
   * <p>Create a module sub-process and configure its listener to accepts messages.</p>
   *
   * @param moduleName
   * @param moduleOpts
   * @param logger
   * @returns {ProcessDispatcher~onReadyCallback}
   */


  _createClass(ProcessDispatcher, [{
    key: "updateRRId",


    /**
     * <p>Creates round robin ids for preforked subprocesses. See how {@link ProcessDispatcher.dispatch} works.</p>
     */
    value: function updateRRId() {
      this._rrIds = (0, _utils.range)(this.availableProcesses.length);
    }
  }, {
    key: "getNextProcId",


    /**
     * <p>Implementation of FIFO queue for selecting of pre-forked subprocess ids.</p>
     * <p>See how {@link ProcessDispatcher.dispatch} works in this regard.</p>
     *
     * @returns {Number} Number of sub-process ID.
     */
    value: function getNextProcId() {
      var nextId = this._rrIds.shift();
      this._rrIds.push(nextId);
      return nextId;
    }
  }, {
    key: "preFork",


    /**
     * <p>Pre-fork subprocess and update a list of available sub-process ids.</p>
     *
     * @param num
     * @returns {ProcessDispatcher~onReadyCallback} On execution finished.
     */
    value: function preFork(num) {
      var _this = this;

      this.logger.debug("preFork " + num);
      return function (callback) {
        async.series((0, _utils.range)(num).map(function () {
          return function (callback) {
            return ProcessDispatcher.makeModuleProcess(_this.moduleName, _this.moduleOpts, _this.logger)(callback);
          };
        }), function (error, moduleProcesses) {
          _this.logger.debug("preFork created " + moduleProcesses.length);
          if (!error && moduleProcesses) {
            _this.availableProcesses = moduleProcesses;
            _this.updateRRId();
            callback();
          } else {
            callback(error);
          }
        });
      };
    }
  }, {
    key: "dispatch",


    /**
     * <p>Dispatch arbitrary function call to available {@link ModuleProcess} instance.</p>
     * <p>The {@link ModuleProcess} selection algorithm is round robin based. See {@link getNextProcId} in this regard.</p>
     *
     * @param functionName  Function name to be called. The desired module wrapped by {@link ModuleProcess} should expose
     *                      API via {@link ProcessDispatcher.listenIPCMessages}. The functionName should be a part of that
     *                      API.
     * @param params        The parameters to be passed to the functionName
     * @returns {ProcessDispatcher~onReadyCallback} On execution finished.
     */
    value: function dispatch(functionName, params) {
      var _this2 = this;

      //this.logger.debug("ProcessDispatcher.dispatch t: ", process.hrtime()[1]);

      var moduleProcess = void 0;

      var availableProcessNum = this.availableProcesses.length;

      if (availableProcessNum) {
        var processNumber = this.getNextProcId();

        this.logger.debug("Dispatching '" + functionName + "' to process number " + processNumber + ". Available processes number " + availableProcessNum);

        moduleProcess = this.availableProcesses[processNumber];
      }

      return function (callback) {
        async.waterfall([function (callback) {
          // Make a descicion on create new process or use existing one.
          if (moduleProcess) {
            _this2.logger.debug("Make use of existing process for '" + _this2.moduleName + "' invocation of '" + functionName + "'");
            callback(null, moduleProcess);
          } else {
            // let errorMessage = "No '" + this.moduleName + "' processes available yet for invocation of '" + functionName + "'";
            // this.logger.debug(errorMessage);
            // callback(new Error(errorMessage));
            _this2.logger.debug("Creating new process for '" + _this2.moduleName + "' invocation of '" + functionName + "'");
            ProcessDispatcher.makeModuleProcess(_this2.moduleName, _this2.moduleOpts, _this2.logger)(callback);
          }
        }, function (moduleProcess, callback) {
          ProcessDispatcher.dispatchToModuleProcess(moduleProcess, functionName, params)(callback);
        }], callback);
      };
    }
  }], [{
    key: "makeModuleProcess",
    value: function makeModuleProcess(moduleName, moduleOpts, logger) {
      return function (callback) {
        return new _moduleProcess.ModuleProcess(moduleName, moduleOpts, logger).init(callback);
      };
    }
  }, {
    key: "dispatchToModule",


    /**
     * <p>Dispatch function invocation to existing moduleProcess.</p>
     *
     * @param moduleName {string}     A js filename containing module function (!) (e.g. module.exports = function() { }).
     *                                I.e. module that is defined by function.
     * @param moduleOpts {Object}     Will be passed to module-function to create a module instance.
     * @param logger
     * @param functionName {string}   Function to be invoked (should be defined on the module).
     * @param params {Object}         Parameters mapping.
     * @returns {ProcessDispatcher~onReadyCallback} On execution finished.
     */
    value: function dispatchToModule(moduleName, moduleOpts, logger, functionName, params) {
      return function (callback) {
        async.waterfall([function (callback) {
          ProcessDispatcher.makeModuleProcess(moduleName, moduleOpts, logger)(callback);
        }, function (moduleProcess, callback) {
          ProcessDispatcher.dispatchToModuleProcess(moduleProcess, functionName, params)(callback);
        }], callback);
      };
    }
  }, {
    key: "dispatchToModuleProcess",


    /**
     * <p>Dispatch function invocation to existing moduleProcess.</p>
     *
     * @param moduleProcess
     * @param functionName {string}   Function defined on the module.
     * @param params {Object}         Parameters mapping.
     * @returns {ProcessDispatcher~onReadyCallback} On execution finished.
     */
    value: function dispatchToModuleProcess(moduleProcess, functionName, params) {
      return function (callback) {
        return moduleProcess.invoke(functionName, params)(callback);
      };
    }
  }, {
    key: "listenIPCMessages",


    /**
     * <p>Create listener to catch the IPC messages within the process the module is running in.</p>
     *
     * @deprecated  This is no longer a part of {@link ProcessDispatcher}. Migrated to
     * {@link ModuleProcess.listenIPCMessages}.
     */
    value: function listenIPCMessages(moduleFn, moduleFileName, logger) {
      _moduleProcess.ModuleProcess.listenIPCMessages(moduleFn, moduleFileName, logger);
    }
  }]);

  return ProcessDispatcher;
}();